import numpy as np
from tqdm import trange

import modules.scripts as scripts
import gradio as gr

from modules import processing, shared, sd_samplers, images
from modules.processing import Processed
from modules.sd_samplers import samplers
from modules.shared import opts, cmd_opts, state

#necessary imports for compositing
from PIL import Image
import random
import cv2

class Script(scripts.Script):
    def title(self):
        return "Double Diffusion v.02"

    def show(self, is_img2img):
        return is_img2img

    def ui(self, is_img2img):
        numcircles = gr.Slider(minimum=0, maximum=1000, step=1, label='Number of Alpha Circles', value=80)
        minradw = gr.Slider(minimum=1, maximum=500, step=1, label='Min radius width', value=60)
        maxradw = gr.Slider(minimum=1, maximum=500, step=1, label='Max radius width', value=120)
        minradh = gr.Slider(minimum=1, maximum=500, step=1, label='Min radius height', value=8)
        maxradh = gr.Slider(minimum=1, maximum=500, step=1, label='Max radius height', value=30)
        minalpha = gr.Slider(minimum=0, maximum=1, step=.01, label='Min Alpha', value=.8)
        maxalpha = gr.Slider(minimum=0, maximum=1, step=.01, label='Max Alpha', value=1.)
        blurrad = gr.Slider(minimum=0.0, maximum=100.0, step=0.1, label='Blur Radius', value=17.0)
        eyeprob = gr.Slider(minimum=0.0, maximum=1.0, step=0.1, label='Eye preservation probability', value=.8)
        mouthprob = gr.Slider(minimum=0.0, maximum=1.0, step=0.1, label='Mouth preservation probability', value=.6)
        second_denoise_strength = gr.Slider(minimum=0.0, maximum=1.0, step=0.01, label='Second Pass Denoising Strength', value=.4)
        secondsteps = gr.Slider(minimum=1, maximum=150, step=1, label='Second Pass Steps', value=100)

        return [numcircles, minradw, maxradw, minradh, maxradh, minalpha, maxalpha, blurrad, eyeprob, mouthprob, second_denoise_strength, secondsteps]

    def run(self, p, numcircles, minradw, maxradw, minradh, maxradh, minalpha, maxalpha, blurrad, eyeprob, mouthprob, second_denoise_strength, secondsteps):
        eye_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_eye.xml')
        mouth_cascade = cv2.CascadeClassifier("C:/ai/haarcascade_mcs_mouth.xml")

        def blend_two(im, secondimg):
            # Load the image
            orig_im = im
            blend_image = secondimg
            blend_image = blend_image.resize(im.size, Image.ANTIALIAS)
            blursize = (int(blurrad), int(blurrad))

            #convert input image to BGR
            face_detect_image = cv2.cvtColor(np.array(im), cv2.COLOR_RGB2BGR)
            image = np.array(blend_image)

            #draw ellipses
            ellis = np.ones(image.shape)
            
            height, width = face_detect_image.shape[:2]
            #Draw i randomly placed and sized ellipses
            for i in range(int(numcircles)):
                # Generate random center point, axes lengths, and angle of rotation for the ellipse
                center = (np.random.randint(0, width), np.random.randint(0, height))
                axes = (np.random.randint(int(minradw), int(maxradw)), np.random.randint(int(minradh), int(maxradh)))
                angle = random.uniform(0.0, 180.0)
                colval = random.uniform(minalpha, maxalpha)
                cv2.ellipse(ellis, center, axes, angle, 0, 360, colval, -1)

            # Draw ellipses around eyes
            eyes = eye_cascade.detectMultiScale(face_detect_image, 1.3, 5)
            for (ex, ey, ew, eh) in eyes:
                if random.random() < eyeprob:
                    cv2.ellipse(ellis, (ex + ew // 2, ey + eh // 2), (ew // 2, eh // 3), 0, 0, 360, (0., 0., 0.), -1)

            # Draw ellipses around mouth(s)
            mouths = mouth_cascade.detectMultiScale(face_detect_image, 1.3, 5)
            for (ex, ey, ew, eh) in mouths:
                if random.random() < mouthprob:
                    cv2.ellipse(ellis, (ex + ew // 2, ey + eh // 2), (ew // 2, eh // 3), 0, 0, 360, (0., 0., 0.), -1)

            #postprocess mask (blur and multiply to convert to 3x8bit img)
            ellis = cv2.blur(ellis, blursize)
            ellis = np.multiply(ellis, 255.0)

            tempimg = Image.fromarray(ellis.astype(np.uint8))
            mask = tempimg.convert("L")

            blend_image.putalpha(mask)
            im.putalpha(255)

            outimg = Image.alpha_composite(im, blend_image)
            return outimg


        processing.fix_seed(p)
        batch_count = p.n_iter
        p.extra_generation_params = {
            "Number of Circles": numcircles,
            "Min Radius Width": minradw,
            "Max Radius Width": maxradw,
            "Min Radius Height": minradh,
            "Max Radius Height": maxradh,
            "Min Alpha": minalpha,
            "Max Alpha": maxalpha,
            "Blur Radius": blurrad,
            "Eye Probability": eyeprob,
            "Mouth Probability": mouthprob,
            "Second Pass Denoise Strength": second_denoise_strength,
            "Second Pass Steps": secondsteps
        }

        p.batch_size = 1
        p.n_iter = 1

        output_images, info = None, None
        initial_seed = None
        initial_info = None

        grids = []
        all_images = []
        
        state.job_count = batch_count * 2
        initial_steps = p.steps
        initial_denoising_strength = p.denoising_strength
        initial_color_corrections = [processing.setup_color_correction(p.init_images[0])]
        original_init_image = p.init_images
        initial_imgs = []
        random.seed(p.seed)

        #--------
        for n in range(batch_count):
            history = []
            p.steps = initial_steps
            # Reset to original init image at the start of each batch
            p.init_images = original_init_image
            p.denoising_strength = initial_denoising_strength
            initial_imgs.append(original_init_image[0])

            p.n_iter = 1
            p.batch_size = 1
            p.do_not_save_grid = True

            if opts.img2img_color_correction:
                p.color_corrections = initial_color_corrections

            processed = processing.process_images(p)

            if initial_seed is None:
                initial_seed = processed.seed
                initial_info = processed.info

            p.seed = processed.seed + 1
            processed.images = processed.images[0]

            history.append(processed.images)

            initme = []
            initimg = blend_two(initial_imgs[0], processed.images)
            initme.append(initimg)
            p.init_images = initme
            p.denoising_strength = second_denoise_strength

            p.steps = secondsteps
            processed = processing.process_images(p)
            if initial_seed is None:
                initial_seed = processed.seed
                initial_info = processed.info

            p.seed = processed.seed + 1
            
            processed.images = processed.images[0]
            history.append(processed.images)

            grid = images.image_grid(history, rows=1)
            if opts.grid_save:
                images.save_image(grid, p.outpath_grids, "grid", initial_seed, p.prompt, opts.grid_format, info=info, short_filename=not opts.grid_extended_filename, grid=True, p=p)

            grids.append(grid)
            all_images += history
            p.init_images = original_init_image

        if opts.return_grid:
            all_images = grids + all_images


        processed = Processed(p, all_images, initial_seed, initial_info)

        return processed